import pytest

from bsv.hd.bip32 import master_xprv_from_seed, bip32_derive_xprvs_from_mnemonic, bip32_derive_xkeys_from_xkey
from bsv.hd.bip39 import seed_from_mnemonic
from bsv.hd.bip44 import bip44_derive_xprvs_from_mnemonic

from bsv.constants import BIP32_DERIVATION_PATH, BIP44_DERIVATION_PATH

# BIP32_DERIVATION_PATH = "m/"
# BIP44_DERIVATION_PATH = "m/44'/236'/0'"

def test_key_derivation_consistency():
    # Test mnemonic phrase
    test_mnemonic = "skin index hair zone brush soldier airport found stuff rare wonder physical"

    # Generate seed from mnemonic
    seed = seed_from_mnemonic(test_mnemonic, lang='en')

    # Generate master keys
    master_xprv = master_xprv_from_seed(seed)
    master_xpub = master_xprv.xpub()

    # Key derivation using different methods
    # 1. BIP32 derivation from master extended private key
    keys_from_bip32_xprv = bip32_derive_xkeys_from_xkey(master_xprv, 0, 2, BIP32_DERIVATION_PATH, 0)
    # 2. BIP32 derivation from master extended public key
    keys_from_bip32_xpub = bip32_derive_xkeys_from_xkey(master_xpub, 0, 2, BIP32_DERIVATION_PATH, 0)
    # 3. BIP32 derivation directly from mnemonic
    keys_from_bip32_mnemonic = bip32_derive_xprvs_from_mnemonic(test_mnemonic, 0, 2, path=BIP32_DERIVATION_PATH, change=0)

    # Test BIP32 derivation consistency
    for i in range(2):
        assert keys_from_bip32_xprv[i].address() == keys_from_bip32_xpub[i].address(), \
            f"BIP32 xprv/xpub derivation mismatch at index {i}"
        assert keys_from_bip32_xprv[i].address() == keys_from_bip32_mnemonic[i].address(), \
            f"BIP32 xprv/mnemonic derivation mismatch at index {i}"

    # Test BIP44 derivation
    keys_from_bip32_mnemonic = bip32_derive_xprvs_from_mnemonic(test_mnemonic, 0, 2, path=BIP44_DERIVATION_PATH, change=0)
    keys_from_bip44_mnemonic = bip44_derive_xprvs_from_mnemonic(test_mnemonic, 0, 2, path=BIP44_DERIVATION_PATH, change=0)

    # Test BIP44 derivation consistency
    for i in range(2):
        assert keys_from_bip32_mnemonic[i].address() == keys_from_bip44_mnemonic[i].address(), \
            f"BIP32/BIP44 derivation mismatch at index {i}"

def test_invalid_mnemonic():
    with pytest.raises(ValueError):
        invalid_mnemonic = "invalid mnemonic phrase"
        bip32_derive_xprvs_from_mnemonic(invalid_mnemonic, 0, 2, path=BIP32_DERIVATION_PATH, change=0)

def test_invalid_derivation_path():
    test_mnemonic = "skin index hair zone brush soldier airport found stuff rare wonder physical"
    with pytest.raises(ValueError):
        invalid_path = "m/invalid"
        bip32_derive_xprvs_from_mnemonic(test_mnemonic, 0, 2, path=invalid_path, change=0)


import hashlib
import hmac
from base64 import b64encode, b64decode
from typing import Optional, Union, Callable, Tuple

from coincurve import PrivateKey as CcPrivateKey, PublicKey as CcPublicKey

from .aes_cbc import aes_decrypt_with_iv
from .aes_cbc import aes_encrypt_with_iv
from .base58 import base58check_encode
from .constants import Network, NETWORK_ADDRESS_PREFIX_DICT, NETWORK_WIF_PREFIX_DICT, PUBLIC_KEY_COMPRESSED_PREFIX_LIST
from .curve import Point
from .curve import curve, curve_multiply as curve_multiply, curve_add as curve_add
from .hash import hash160, hash256, hmac_sha256
from .utils import decode_wif, text_digest, stringify_ecdsa_recoverable, unstringify_ecdsa_recoverable
from .utils import deserialize_ecdsa_recoverable, serialize_ecdsa_der
from .polynomial import Polynomial, PointInFiniteField, KeyShares


class PublicKey:

    def __init__(self, public_key: Union[str, bytes, Point, CcPublicKey]):
        """
        create public key from serialized hex string or bytes, or curve point, or CoinCurve public key
        """
        self.compressed: bool = True  # use compressed format public key by default
        if isinstance(public_key, Point):
            # from curve point
            self.key: CcPublicKey = CcPublicKey.from_point(public_key.x, public_key.y)
        elif isinstance(public_key, CcPublicKey):
            # from CoinCurve public key
            self.key: CcPublicKey = public_key
        else:
            if isinstance(public_key, str):
                # from serialized public key in hex string
                pk: bytes = bytes.fromhex(public_key)
            elif isinstance(public_key, bytes):
                # from serialized public key in bytes
                pk: bytes = public_key
            else:
                raise TypeError('unsupported public key type')
            # here we have serialized public key in bytes
            self.key: CcPublicKey = CcPublicKey(pk)
            self.compressed: bool = pk[:1] in PUBLIC_KEY_COMPRESSED_PREFIX_LIST

    def point(self) -> Point:
        return Point(*self.key.point())

    def serialize(self, compressed: Optional[bool] = None) -> bytes:
        compressed = self.compressed if compressed is None else compressed
        return self.key.format(compressed)

    def hex(self, compressed: Optional[bool] = None) -> str:
        return self.serialize(compressed).hex()

    def hash160(self, compressed: Optional[bool] = None) -> bytes:
        """
        :returns: public key hash corresponding to this public key
        """
        return hash160(self.serialize(compressed))

    hash = hash160

    def address(self, compressed: Optional[bool] = None, network: Network = Network.MAINNET) -> str:
        """
        :returns: P2PKH address corresponding to this public key
        """
        return base58check_encode(NETWORK_ADDRESS_PREFIX_DICT.get(network) + self.hash160(compressed))

    def verify(self, signature: bytes, message: bytes, hasher: Optional[Callable[[bytes], bytes]] = hash256) -> bool:
        """
        verify serialized ECDSA signature in bitcoin strict DER (low-s) format
        """
        return self.key.verify(signature, message, hasher)

    def verify_recoverable(
            self,
            signature: bytes,
            message: bytes,
            hasher: Optional[Callable[[bytes], bytes]] = hash256
    ) -> bool:
        """
        verify serialized recoverable ECDSA signature in format "r (32 bytes) + s (32 bytes) + recovery_id (1 byte)"
        """
        r, s, _ = deserialize_ecdsa_recoverable(signature)
        der = serialize_ecdsa_der((r, s))
        return self.verify(der, message, hasher) and self == recover_public_key(signature, message, hasher)

    def derive_shared_secret(self, key: 'PrivateKey') -> bytes:
        return PublicKey(self.key.multiply(key.serialize())).serialize()

    def encrypt(self, message: bytes) -> bytes:
        """
        Electrum ECIES (aka BIE1) encryption
        """
        # generate an ephemeral EC private key in order to derive shared secret (ECDH key)
        ephemeral_private_key = PrivateKey()
        # derive ECDH key
        ecdh_key: bytes = self.derive_shared_secret(ephemeral_private_key)
        # SHA512(ECDH_KEY), then we have
        # key_e and iv used in AES, key_m used in HMAC.SHA256
        key: bytes = hashlib.sha512(ecdh_key).digest()
        iv, key_e, key_m = key[0:16], key[16:32], key[32:]
        # make AES encryption
        cipher: bytes = aes_encrypt_with_iv(key_e, iv, message)
        # encrypted = magic_bytes (4 bytes) + ephemeral_public_key (33 bytes) + cipher (16 bytes at least)
        encrypted: bytes = 'BIE1'.encode('utf-8') + ephemeral_private_key.public_key().serialize() + cipher
        # mac = HMAC_SHA256(encrypted), 32 bytes
        mac: bytes = hmac.new(key_m, encrypted, hashlib.sha256).digest()
        # give out encrypted + mac
        return encrypted + mac

    def encrypt_text(self, text: str) -> str:
        """
        :returns: BIE1 encrypted text, base64 encoded
        """
        message: bytes = text.encode('utf-8')
        return b64encode(self.encrypt(message)).decode('ascii')

    def derive_child(self, private_key: 'PrivateKey', invoice_number: str) -> 'PublicKey':
        """
        derive a child key with BRC-42
        :param private_key: the private key of the other party
        :param invoice_number: the invoice number used to derive the child key
        :return: the derived child key
        """
        shared_key = self.derive_shared_secret(private_key)
        hashing = hmac_sha256(shared_key, invoice_number.encode('utf-8'))
        point = curve_multiply(int.from_bytes(hashing, 'big'), curve.g)
        final_point = curve_add(self.point(), point)
        return PublicKey(final_point)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, PublicKey):
            return self.key == o.key
        return super().__eq__(o)  # pragma: no cover

    def __str__(self) -> str:  # pragma: no cover
        return f'<PublicKey hex={self.hex()}>'

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()


class PrivateKey:

    def __init__(
            self,
            private_key: Union[str, int, bytes, CcPrivateKey, None] = None,
            network: Optional[Network] = None
    ):
        """
        create private key from WIF (str), or int, or bytes, or CoinCurve private key
        random a new private key if None
        """
        self.network: Network = network or Network.MAINNET
        self.compressed: bool = True  # use compressed WIF by default
        if private_key is None:
            # create a new private key
            self.key: CcPrivateKey = CcPrivateKey()
        elif isinstance(private_key, CcPrivateKey):
            # from CoinCurve private key
            self.key: CcPrivateKey = private_key
        else:
            if isinstance(private_key, str):
                # from wif
                private_key_bytes, self.compressed, self.network = decode_wif(private_key)
                self.key: CcPrivateKey = CcPrivateKey(private_key_bytes)
            elif isinstance(private_key, int):
                # from private key as int
                self.key: CcPrivateKey = CcPrivateKey.from_int(private_key)
            elif isinstance(private_key, bytes):
                # from private key integer in bytes
                self.key: CcPrivateKey = CcPrivateKey(private_key)
            else:
                raise TypeError('unsupported private key type')

    def public_key(self) -> PublicKey:
        return PublicKey(self.key.public_key.format(self.compressed))

    def address(self, compressed: Optional[bool] = None, network: Optional[Network] = None) -> str:
        """
        :returns: P2PKH address corresponding to this private key
        """
        compressed = self.compressed if compressed is None else compressed
        network = network or self.network
        return self.public_key().address(compressed, network)

    def wif(self, compressed: Optional[bool] = None, network: Optional[Network] = None) -> str:
        compressed = self.compressed if compressed is None else compressed
        network = network or self.network
        key_bytes = self.serialize()
        compressed_bytes = b'\x01' if compressed else b''
        return base58check_encode(NETWORK_WIF_PREFIX_DICT.get(network) + key_bytes + compressed_bytes)

    def int(self) -> int:
        return self.key.to_int()

    def serialize(self) -> bytes:
        return self.key.secret

    def hex(self) -> str:
        return self.serialize().hex()

    def der(self) -> bytes:  # pragma: no cover
        return self.key.to_der()

    def pem(self) -> bytes:  # pragma: no cover
        return self.key.to_pem()

    def sign(self, message: bytes, hasher: Optional[Callable[[bytes], bytes]] = hash256, k: Optional[int] = None) -> bytes:
        """
        :returns: ECDSA signature in bitcoin strict DER (low-s) format
        """
        if k:
            return self._sign_custom_k(message, hasher, k)
        return self.key.sign(message, hasher)
    
    def _sign_custom_k(self, message: bytes, hasher: Callable[[bytes], bytes], k: int) -> bytes:
        # TODO: This could be done using self.key.sign() but the interface needs a custom k value function to be injected into te C binary
        #       of libsecp256k1, since the default one does some transformations to the value.
        #       See https://github.com/rustyrussell/secp256k1-py/blob/5bad581d959d722bf6c2df5eaa996fd4c24096aa/tests/test_custom_nonce.py#L51ffi%20=%20FFI()
        #           https://github.com/bitcoin-core/secp256k1/blob/master/src/secp256k1.c#L518

        z = int.from_bytes(hasher(message), 'big')

        # Ensure k is valid
        k = k % curve.n
        if k == 0:
            raise ValueError("Invalid nonce k")

        # Compute R = k * G and obtain its x-coordinate (r)
        R = curve_multiply(k, curve.g)
        if R is None:
            raise ValueError("Invalid R value")
        r = R.x

        # Compute s = k^(-1) * (z + r * d) mod n
        d = int.from_bytes(self.serialize(), 'big')
        s = (pow(k, -1, curve.n) * (z + r * d)) % curve.n
        if s == 0:
            raise ValueError("Invalid s value")

        # Ensure the signature is canonical (low S value)
        if s > curve.n // 2:
            s = curve.n - s

        # Convert r and s to bytes
        r_bytes = r.to_bytes(32, 'big')
        s_bytes = s.to_bytes(32, 'big')

        # Add prefix if the MSB is set
        if r_bytes[0] & 0x80:
            r_bytes = b'\x00' + r_bytes
        if s_bytes[0] & 0x80:
            s_bytes = b'\x00' + s_bytes

        # Serialize the signature in DER format
        signature = b'\x30' + (4 + len(r_bytes) + len(s_bytes)).to_bytes(1, 'big') + \
                    b'\x02' + len(r_bytes).to_bytes(1, 'big') + r_bytes + \
                    b'\x02' + len(s_bytes).to_bytes(1, 'big') + s_bytes

        return signature

    def verify(self, signature: bytes, message: bytes, hasher: Optional[Callable[[bytes], bytes]] = hash256) -> bool:
        """
        verify ECDSA signature in bitcoin strict DER (low-s) format
        """
        return self.public_key().verify(signature, message, hasher)

    def sign_recoverable(self, message: bytes, hasher: Optional[Callable[[bytes], bytes]] = hash256) -> bytes:
        """
        :returns: serialized recoverable ECDSA signature (aka compact signature) in format
                    r (32 bytes) + s (32 bytes) + recovery_id (1 byte)
        """
        return self.key.sign_recoverable(message, hasher)

    def verify_recoverable(
            self,
            signature: bytes,
            message: bytes,
            hasher: Optional[Callable[[bytes], bytes]] = hash256
    ) -> bool:
        """
        verify serialized recoverable ECDSA signature in format "r (32 bytes) + s (32 bytes) + recovery_id (1 byte)"
        """
        return self.public_key().verify_recoverable(signature, message, hasher)

    def sign_text(self, text: str) -> Tuple[str, str]:
        """sign arbitrary text with bitcoin private key
        :returns: (p2pkh_address, stringified_recoverable_ecdsa_signature)
        This function follows Bitcoin Signed Message Format.
        For BRC-77, use signed_message.py instead.
        """
        message: bytes = text_digest(text)
        return self.address(), stringify_ecdsa_recoverable(self.sign_recoverable(message), self.compressed)

    def derive_shared_secret(self, key: PublicKey) -> bytes:
        return PublicKey(key.key.multiply(self.serialize())).serialize()

    def decrypt(self, message: bytes) -> bytes:
        """
        Electrum ECIES (aka BIE1) decryption
        """
        assert len(message) >= 85, 'invalid encrypted length'
        encrypted, mac = message[:-32], message[-32:]
        # encrypted = magic_bytes (4 bytes) + ephemeral_public_key (33 bytes) + cipher_text (16 bytes at least)
        magic_bytes, ephemeral_public_key, cipher = encrypted[:4], PublicKey(encrypted[4:37]), encrypted[37:]
        assert magic_bytes.decode('utf-8') == 'BIE1', 'invalid magic bytes'
        # restore ECDH key
        ecdh_key = self.derive_shared_secret(ephemeral_public_key)
        # restore iv, key_e, key_m
        key = hashlib.sha512(ecdh_key).digest()
        iv, key_e, key_m = key[0:16], key[16:32], key[32:]
        # verify mac
        assert hmac.new(key_m, encrypted, hashlib.sha256).digest().hex() == mac.hex(), 'incorrect hmac checksum'
        # make the AES decryption
        return aes_decrypt_with_iv(key_e, iv, cipher)

    def decrypt_text(self, text: str) -> str:
        """
        decrypt BIE1 encrypted, base64 encoded text
        """
        message: bytes = b64decode(text)
        return self.decrypt(message).decode('utf-8')

    def encrypt(self, message: bytes) -> bytes:  # pragma: no cover
        """
        Electrum ECIES (aka BIE1) encryption
        """
        return self.public_key().encrypt(message)

    def encrypt_text(self, text: str) -> str:  # pragma: no cover
        """
        :returns: BIE1 encrypted text, base64 encoded
        """
        return self.public_key().encrypt_text(text)

    def derive_child(self, public_key: PublicKey, invoice_number: str) -> 'PrivateKey':
        """
        derive a child key with BRC-42
        :param public_key: the public key of the other party
        :param invoice_number: the invoice number used to derive the child key
        :return: the derived child key
        """
        shared_key = self.derive_shared_secret(public_key)
        hashing = hmac_sha256(shared_key, invoice_number.encode('utf-8'))
        return PrivateKey((self.int() + int.from_bytes(hashing, 'big')) % curve.n)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, PrivateKey):
            return self.key == o.key
        return super().__eq__(o)  # pragma: no cover

    def __str__(self) -> str:  # pragma: no cover
        return f'<PrivateKey wif={self.wif()} int={self.int()}>'

    def __repr__(self) -> str:  # pragma: no cover
        return self.__str__()

    @classmethod
    def from_hex(cls, octets: Union[str, bytes]) -> 'PrivateKey':
        b: bytes = octets if isinstance(octets, bytes) else bytes.fromhex(octets)
        return PrivateKey(CcPrivateKey(b))

    @classmethod
    def from_der(cls, octets: Union[str, bytes]) -> 'PrivateKey':  # pragma: no cover
        b: bytes = octets if isinstance(octets, bytes) else bytes.fromhex(octets)
        return PrivateKey(CcPrivateKey.from_der(b))

    @classmethod
    def from_pem(cls, octets: Union[str, bytes]) -> 'PrivateKey':  # pragma: no cover
        b: bytes = octets if isinstance(octets, bytes) else bytes.fromhex(octets)
        return PrivateKey(CcPrivateKey.from_pem(b))

    def to_key_shares(self, threshold: int, total_shares: int) -> 'KeyShares':
        """
        Split the private key into shares using Shamir's Secret Sharing Scheme.

        Args:
            threshold: The minimum number of shares required to reconstruct the private key
            total_shares: The total number of shares to generate

        Returns:
            A KeyShares object containing the generated shares

        Raises:
            ValueError: If threshold or total_shares are invalid
        """

        # Input validation
        if not isinstance(threshold, int) or not isinstance(total_shares, int):
            raise ValueError("threshold and totalShares must be numbers")
        if threshold < 2:
            raise ValueError("threshold must be at least 2")
        if total_shares < 2:
            raise ValueError("totalShares must be at least 2")
        if threshold > total_shares:
            raise ValueError("threshold should be less than or equal to totalShares")

        # Create polynomial from private key
        poly = Polynomial.from_private_key(self.int(), threshold)

        # Generate shares
        points = []
        for i in range(total_shares):
            # Generate random x coordinate using a new private key
            # Using private_key.key.to_int() based on the structure in keys.py
            random_private_key = PrivateKey()
            x = random_private_key.int()

            # Evaluate polynomial at x to get y coordinate
            y = poly.value_at(x)

            # Create a point and add to points' list
            points.append(PointInFiniteField(x, y))

        # Calculate integrity hash from the public key
        # In the JS implementation: (this.toPublicKey().toHash('hex') as string).slice(0, 8)
        integrity = self.public_key().hash160().hex()[:8]

        return KeyShares(points, threshold, integrity)

    def to_backup_shares(self, threshold: int, total_shares: int) -> list:
        """
        Creates a backup of the private key by splitting it into shares.

        Args:
            threshold: The number of shares which will be required to reconstruct the private key
            total_shares: The number of shares to generate for distribution

        Returns:
            List of share strings in backup format
        """
        key_shares = self.to_key_shares(threshold, total_shares)
        return key_shares.to_backup_format()

    @staticmethod
    def from_backup_shares(shares: list) -> 'PrivateKey':
        """
        Reconstructs a private key from backup shares.

        Args:
            shares: List of share strings in backup format

        Returns:
            The reconstructed PrivateKey object

        Raises:
            ValueError: If shares are invalid or inconsistent
        """
        return PrivateKey.from_key_shares(KeyShares.from_backup_format(shares))

    @staticmethod
    def from_key_shares(key_shares: 'KeyShares') -> 'PrivateKey':
        """
        Combines shares to reconstruct the private key.

        Args:
            key_shares: A KeyShares object containing the shares

        Returns:
            The reconstructed PrivateKey object

        Raises:
            ValueError: If not enough shares are provided or shares are invalid
        """


        points = key_shares.points
        threshold = key_shares.threshold
        integrity = key_shares.integrity

        # Validate inputs
        if threshold < 2:
            raise ValueError("threshold must be at least 2")
        if len(points) < threshold:
            raise ValueError(f"At least {threshold} shares are required to reconstruct the private key")

        # Check for duplicate x values
        for i in range(threshold):
            for j in range(i + 1, threshold):
                if points[i].x == points[j].x:
                    raise ValueError("Duplicate share detected, each must be unique.")

        # Create polynomial from points
        poly = Polynomial(points[:threshold], threshold)

        # Evaluate polynomial at x=0 to get the private key
        secret_value = poly.value_at(0)

        # Create private key from secret value
        # Instead of from_int (which doesn't exist), use the proper constructor
        private_key = PrivateKey(secret_value)

        # Verify integrity by comparing hash of public key
        reconstructed_integrity = private_key.public_key().hash160().hex()[:8]
        if reconstructed_integrity != integrity:
            raise ValueError("Integrity hash mismatch")

        return private_key

def verify_signed_text(
        text: str,
        address: str,
        signature: str,
        hasher: Optional[Callable[[bytes], bytes]] = hash256
) -> bool:
    """
    verify signed arbitrary text
    """
    serialized_recoverable, compressed = unstringify_ecdsa_recoverable(signature)
    r, s, _ = deserialize_ecdsa_recoverable(serialized_recoverable)
    message: bytes = text_digest(text)
    public_key: PublicKey = recover_public_key(serialized_recoverable, message, hasher)
    der: bytes = serialize_ecdsa_der((r, s))
    return public_key.verify(der, message, hasher) and public_key.address(compressed=compressed) == address


def recover_public_key(
        signature: bytes,
        message: bytes,
        hasher: Optional[Callable[[bytes], bytes]] = hash256
) -> PublicKey:
    """
    recover public key from serialized recoverable ECDSA signature in format
      "r (32 bytes) + s (32 bytes) + recovery_id (1 byte)"
    """
    return PublicKey(CcPublicKey.from_signature_and_message(signature, message, hasher))



import pytest
from bsv.script.script import Script
from bsv.constants import OpCode


def test_script_build_chunks_pushdata_opcodes():
    """
    Test that the Script._build_chunks method correctly handles PUSHDATA opcodes
    when changing the reading method from byte-by-int to unit-based reading.
    """

    # Test PUSHDATA1 with a length value that would be negative if incorrectly interpreted as signed
    # 0xff = 255 bytes of data
    pushdata1_high_length = b'\x4c\xff' + b'\x42' * 255
    script_pushdata1 = Script(pushdata1_high_length)
    assert len(script_pushdata1.chunks) == 1
    assert script_pushdata1.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_pushdata1.chunks[0].data == b'\x42' * 255
    assert len(script_pushdata1.chunks[0].data) == 255

    # Test with smaller data sizes to ensure consistent behavior
    pushdata1_75 = b'\x4c\xff' + b'\x42' * 75
    script_pushdata1_75 = Script(pushdata1_75)
    assert len(script_pushdata1_75.chunks) == 1
    assert script_pushdata1_75.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_pushdata1_75.chunks[0].data == b'\x42' * 75

    pushdata1_76 = b'\x4c\xff' + b'\x42' * 76
    script_pushdata1_76 = Script(pushdata1_76)
    assert len(script_pushdata1_76.chunks) == 1
    assert script_pushdata1_76.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_pushdata1_76.chunks[0].data == b'\x42' * 76

    # Test PUSHDATA2 with a length value that would be negative if incorrectly interpreted as signed
    # 0xffff = 65535 bytes of data
    pushdata2_high_length = b'\x4d\xff\xff' + b'\x42' * 65535
    script_pushdata2 = Script(pushdata2_high_length)
    assert len(script_pushdata2.chunks) == 1
    assert script_pushdata2.chunks[0].op == OpCode.OP_PUSHDATA2
    assert script_pushdata2.chunks[0].data == b'\x42' * 65535
    assert len(script_pushdata2.chunks[0].data) == 65535

    # Test with smaller data sizes for PUSHDATA2
    pushdata2_255 = b'\x4d\xff\xff' + b'\x42' * 255
    script_pushdata2_255 = Script(pushdata2_255)
    assert len(script_pushdata2_255.chunks) == 1
    assert script_pushdata2_255.chunks[0].op == OpCode.OP_PUSHDATA2
    assert script_pushdata2_255.chunks[0].data == b'\x42' * 255

    pushdata2_256 = b'\x4d\xff\xff' + b'\x42' * 256
    script_pushdata2_256 = Script(pushdata2_256)
    assert len(script_pushdata2_256.chunks) == 1
    assert script_pushdata2_256.chunks[0].op == OpCode.OP_PUSHDATA2
    assert script_pushdata2_256.chunks[0].data == b'\x42' * 256

    # Test PUSHDATA4 with values that would be negative if interpreted as signed integers
    # Test with very large value - 0x80000001 = 2,147,483,649 (would be -2,147,483,647 as signed int32)
    # Note: This test may require significant memory
    pushdata4_large_value = b'\x4e\x01\x00\x00\x80' + b'\x42' * 2147483649
    script_pushdata4_large = Script(pushdata4_large_value)
    assert len(script_pushdata4_large.chunks) == 1
    assert script_pushdata4_large.chunks[0].op == OpCode.OP_PUSHDATA4
    assert len(script_pushdata4_large.chunks[0].data) == 2147483649

    # Test with smaller data sizes for PUSHDATA4
    pushdata4_upper_half = b'\x4e\x00\x00\x00\xC0' + b'\x43' * 65535
    script_pushdata4_upper_half = Script(pushdata4_upper_half)
    assert len(script_pushdata4_upper_half.chunks) == 1
    assert script_pushdata4_upper_half.chunks[0].op == OpCode.OP_PUSHDATA4
    assert len(script_pushdata4_upper_half.chunks[0].data) == 65535

    # Test with slightly larger data size
    pushdata4_upper_half_2 = b'\x4e\x00\x00\x00\xC0' + b'\x43' * 65536
    script_pushdata4_upper_half_2 = Script(pushdata4_upper_half_2)
    assert len(script_pushdata4_upper_half_2.chunks) == 1
    assert script_pushdata4_upper_half_2.chunks[0].op == OpCode.OP_PUSHDATA4
    assert len(script_pushdata4_upper_half_2.chunks[0].data) == 65536

    # Test boundary cases where the length is exactly at important thresholds
    # PUSHDATA1 with length 0
    pushdata1_zero = b'\x4c\x00'
    script_pushdata1_zero = Script(pushdata1_zero)
    assert len(script_pushdata1_zero.chunks) == 1
    assert script_pushdata1_zero.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_pushdata1_zero.chunks[0].data == b''
    assert len(script_pushdata1_zero.chunks[0].data) == 0

    # Edge case: PUSHDATA with incomplete length specification
    incomplete_pushdata1 = b'\x4c'  # PUSHDATA1 without length byte
    script_incomplete1 = Script(incomplete_pushdata1)
    assert len(script_incomplete1.chunks) == 1
    assert script_incomplete1.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_incomplete1.chunks[0].data is None

    incomplete_pushdata2 = b'\x4d\xff'  # PUSHDATA2 with incomplete length (only one byte)
    script_incomplete2 = Script(incomplete_pushdata2)
    assert len(script_incomplete2.chunks) == 1
    assert script_incomplete2.chunks[0].op == OpCode.OP_PUSHDATA2
    assert script_incomplete2.chunks[0].data == b''

    # Edge case: PUSHDATA with specified length but insufficient data
    insufficient_data1 = b'\x4c\x0A\x01\x02\x03'  # PUSHDATA1 expecting 10 bytes but only 3 are provided
    script_insufficient1 = Script(insufficient_data1)
    assert len(script_insufficient1.chunks) == 1
    assert script_insufficient1.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_insufficient1.chunks[0].data == b'\x01\x02\x03'  # Should get the available data

    # Multiple PUSHDATA opcodes in sequence to test parsing continuity
    mixed_pushdata = (
        b'\x4c\x03\x01\x02\x03'  # PUSHDATA1 with 3 bytes
        b'\x4d\x04\x00\x04\x05\x06\x07'  # PUSHDATA2 with 4 bytes
        b'\x02\x08\x09'  # Direct push of 2 bytes
    )
    script_mixed = Script(mixed_pushdata)
    assert len(script_mixed.chunks) == 3
    assert script_mixed.chunks[0].op == OpCode.OP_PUSHDATA1
    assert script_mixed.chunks[0].data == b'\x01\x02\x03'
    assert script_mixed.chunks[1].op == OpCode.OP_PUSHDATA2
    assert script_mixed.chunks[1].data == b'\x04\x05\x06\x07'
    assert script_mixed.chunks[2].op == b'\x02'
    assert script_mixed.chunks[2].data == b'\x08\x09'


def test_script_serialization_with_pushdata():
    """
    Test that serialization and deserialization of scripts with PUSHDATA opcodes work correctly.

    This test verifies that scripts containing PUSHDATA opcodes can be:
    1. Serialized back to their original binary form
    2. Deserialized from binary to produce identical Script objects with properly parsed chunks

    This ensures the round-trip integrity of Script objects with various PUSHDATA operations.
    """
    # Create a script with various PUSHDATA opcodes and direct push data
    original_script = (
        b'\x4c\x03\x01\x02\x03'  # PUSHDATA1 with 3 bytes
        b'\x4d\x04\x00\x04\x05\x06\x07'  # PUSHDATA2 with 4 bytes
        b'\x02\x08\x09'  # Direct push of 2 bytes
    )

    script = Script(original_script)

    # Serialize and deserialize the script
    serialized = script.serialize()
    deserialized = Script(serialized)

    # Verify the scripts are equivalent
    assert serialized == original_script
    assert deserialized.serialize() == original_script

    # Check that the chunks are correctly parsed
    assert len(deserialized.chunks) == 3
    assert deserialized.chunks[0].op == OpCode.OP_PUSHDATA1
    assert deserialized.chunks[0].data == b'\x01\x02\x03'
    assert deserialized.chunks[1].op == OpCode.OP_PUSHDATA2
    assert deserialized.chunks[1].data == b'\x04\x05\x06\x07'
    assert deserialized.chunks[2].op == b'\x02'
    assert deserialized.chunks[2].data == b'\x08\x09'


if __name__ == "__main__":
    test_script_build_chunks_pushdata_opcodes()
    test_script_serialization_with_pushdata()
    print("All tests passed!")



import math
from contextlib import suppress
from typing import List, Optional, Union, Dict, Any

from .broadcaster import Broadcaster, BroadcastResponse
from .broadcasters import default_broadcaster
from .chaintracker import ChainTracker
from .chaintrackers import default_chain_tracker
from .fee_models import SatoshisPerKilobyte
from .constants import (
    TRANSACTION_VERSION,
    TRANSACTION_LOCKTIME,
    TRANSACTION_FEE_RATE,
)
from .hash import hash256
from .merkle_path import MerklePath
from .script.spend import Spend
from .script.type import P2PKH
from .transaction_input import TransactionInput
from .transaction_output import TransactionOutput
from .transaction_preimage import tx_preimage
from .utils import unsigned_to_varint, Reader, Writer, reverse_hex_byte_order


class InsufficientFunds(ValueError):
    pass


class Transaction:

    def __init__(
            self,
            tx_inputs: Optional[List[TransactionInput]] = None,
            tx_outputs: Optional[List[TransactionOutput]] = None,
            version: int = TRANSACTION_VERSION,
            locktime: int = TRANSACTION_LOCKTIME,
            merkle_path: Optional[MerklePath] = None,
            **kwargs,
    ):
        self.inputs: List[TransactionInput] = tx_inputs or []
        self.outputs: List[TransactionOutput] = tx_outputs or []
        self.version: int = version
        self.locktime: int = locktime
        self.merkle_path = merkle_path

        self.kwargs: Dict[str, Any] = dict(**kwargs) or {}

    def serialize(self) -> bytes:
        raw = self.version.to_bytes(4, "little")
        raw += unsigned_to_varint(len(self.inputs))
        for tx_input in self.inputs:
            raw += tx_input.serialize()
        raw += unsigned_to_varint(len(self.outputs))
        for tx_output in self.outputs:
            raw += tx_output.serialize()
        raw += self.locktime.to_bytes(4, "little")
        return raw

    def add_input(
            self, tx_input: TransactionInput
    ) -> "Transaction":  # pragma: no cover
        if isinstance(tx_input, TransactionInput):
            self.inputs.append(tx_input)
        else:
            raise TypeError("unsupported transaction input type")
        return self

    def add_inputs(self, tx_inputs: List[TransactionInput]) -> "Transaction":
        for tx_input in tx_inputs:
            self.add_input(tx_input)
        return self

    def add_output(self, tx_output: TransactionOutput) -> "Transaction":  # pragma: no cover
        self.outputs.append(tx_output)
        return self

    def add_outputs(self, tx_outputs: List[TransactionOutput]) -> "Transaction":
        for tx_output in tx_outputs:
            self.add_output(tx_output)
        return self

    def hex(self) -> str:  # pragma: no cover
        return self.serialize().hex()

    raw = hex

    def hash(self) -> bytes:
        return hash256(self.serialize())

    def txid(self) -> str:
        return self.hash()[::-1].hex()

    def preimage(self, index: int) -> bytes:
        """
        :returns: digest of the input specified by index
        """
        assert (
                0 <= index < len(self.inputs)
        ), f"index out of range [0, {len(self.inputs)})"
        return tx_preimage(index, self.inputs, self.outputs, self.version, self.locktime)

    def sign(self, bypass: bool = True) -> "Transaction":  # pragma: no cover
        """
        :bypass: if True then ONLY sign inputs which unlocking script is None, otherwise sign all the inputs
        sign all inputs according to their script type
        """
        for out in self.outputs:
            if out.satoshis is None:
                if out.change:
                    raise ValueError('There are still change outputs with uncomputed amounts. Use the fee() method to compute the change amounts and transaction fees prior to signing.')
                else:
                    raise ValueError('One or more transaction outputs is missing an amount. Ensure all output amounts are provided before signing.')

        for i in range(len(self.inputs)):
            tx_input = self.inputs[i]
            if tx_input.unlocking_script is None or not bypass:
                tx_input.unlocking_script = tx_input.unlocking_script_template.sign(
                    self, i
                )
        return self

    def total_value_in(self) -> int:
        return sum([tx_input.satoshis for tx_input in self.inputs])

    def total_value_out(self) -> int:
        return sum([tx_output.satoshis for tx_output in self.outputs])

    def get_fee(self) -> int:
        """
        :returns: actual fee paid of this transaction under the current state
        """
        return self.total_value_in() - self.total_value_out()

    def byte_length(self) -> int:
        """
        :returns: actual byte length of this transaction under the current state
        """
        return len(self.serialize())

    size = byte_length

    def estimated_byte_length(self) -> int:
        """
        :returns: estimated byte length of this transaction after signing
        if transaction has already signed, it will return the same value as function byte_length
        """
        estimated_length = (
                4
                + len(unsigned_to_varint(len(self.inputs)))
                + len(unsigned_to_varint(len(self.outputs)))
                + 4
        )
        for tx_input in self.inputs:
            if tx_input.unlocking_script is not None:
                # unlocking script already set
                estimated_length += len(tx_input.serialize())
            else:
                estimated_length += (
                        41
                        + tx_input.unlocking_script_template.estimated_unlocking_byte_length()
                )
        for tx_output in self.outputs:
            estimated_length += (
                    8
                    + len(tx_output.locking_script.byte_length_varint())
                    + tx_output.locking_script.byte_length()
            )
        return estimated_length

    estimated_size = estimated_byte_length

    def fee(self, model_or_fee=None, change_distribution='equal'):
        """
        Computes the fee for the transaction and adjusts the change outputs accordingly.
        
        :param model_or_fee: Fee model or fee amount. Defaults to `SatoshisPerKilobyte` with value 10 if not provided.
        :param change_distribution: Method of change distribution ('equal' or 'random'). Defaults to 'equal'.
        """
        
        if model_or_fee is None:
            model_or_fee = SatoshisPerKilobyte(int(TRANSACTION_FEE_RATE))

        if isinstance(model_or_fee, int):
            fee = model_or_fee
        else:
            fee = model_or_fee.compute_fee(self)

        change = 0
        for tx_in in self.inputs:
            if not tx_in.source_transaction:
                raise ValueError('Source transactions are required for all inputs during fee computation')
            change += tx_in.source_transaction.outputs[tx_in.source_output_index].satoshis
        
        change -= fee
        
        change_count = 0
        for out in self.outputs:
            if not out.change:
                change -= out.satoshis
            else:
                change_count += 1
        
        if change <= change_count:
            # Not enough change to distribute among the change outputs.
            # Remove all change outputs and leave the extra for the miners.
            self.outputs = [out for out in self.outputs if not out.change]
            return
        
        # Distribute change among change outputs
        if change_distribution == 'random':
            # TODO: Implement random distribution
            raise NotImplementedError('Random change distribution is not yet implemented')
        elif change_distribution == 'equal':
            per_output = change // change_count
            for out in self.outputs:
                if out.change:
                    out.satoshis = per_output

    async def broadcast(
            self,
            broadcaster: Broadcaster = default_broadcaster(),
            check_fee: bool = True
    ) -> BroadcastResponse:  # pragma: no cover
        return await broadcaster.broadcast(self)

    @classmethod
    def from_hex(cls, stream: Union[str, bytes, Reader]) -> Optional["Transaction"]:
        with suppress(Exception):
            if isinstance(stream, str):
                return cls.from_reader(Reader(bytes.fromhex(stream)))
            elif isinstance(stream, bytes):
                return cls.from_reader(Reader(stream))
            return cls.from_reader(stream)
        return None

    @classmethod
    def from_beef(cls, stream: Union[str, bytes, Reader]) -> "Transaction":
        stream = (
            stream
            if isinstance(stream, Reader)
            else Reader(
                stream if isinstance(stream, bytes) else bytes.fromhex(stream)
            )
        )
        version = stream.read_uint32_le()
        if version != 4022206465:
            raise ValueError(
                f"Invalid BEEF version. Expected 4022206465, received {version}."
            )

        number_of_bumps = stream.read_var_int_num()
        bumps = []
        for _ in range(number_of_bumps):
            bumps.append(MerklePath.from_reader(stream))

        number_of_transactions = stream.read_var_int_num()
        transactions = {}
        last_txid = None
        for i in range(number_of_transactions):
            tx = cls.from_reader(stream)
            obj = {"tx": tx}
            txid = tx.txid()
            if i + 1 == number_of_transactions:
                last_txid = txid
            has_bump = bool(stream.read_uint8())
            if has_bump:
                obj["pathIndex"] = stream.read_var_int_num()
            transactions[txid] = obj

        def add_path_or_inputs(item):
            if "pathIndex" in item:
                path = bumps[item["pathIndex"]]
                if not isinstance(path, MerklePath):
                    raise ValueError("Invalid merkle path index found in BEEF!")
                item["tx"].merkle_path = path
            else:
                for tx_input in item["tx"].inputs:
                    source_obj = transactions[tx_input.source_txid]
                    if not isinstance(source_obj, dict):
                        raise ValueError(
                            f"Reference to unknown TXID in BUMP: {tx_input.source_txid}"
                        )
                    tx_input.source_transaction = source_obj["tx"]
                    add_path_or_inputs(source_obj)

        add_path_or_inputs(transactions[last_txid])
        return transactions[last_txid]["tx"]

    def to_ef(self) -> bytes:
        writer = Writer()
        writer.write_uint32_le(self.version)
        writer.write(bytes.fromhex('0000000000ef'))
        writer.write_var_int_num(len(self.inputs))

        for i in self.inputs:
            if i.source_transaction is None:
                raise ValueError('All inputs must have source transactions when serializing to EF format')
            if i.source_txid and i.source_txid != '00' * 32:
                writer.write(bytes.fromhex(reverse_hex_byte_order(i.source_txid)))
            else:
                writer.write(i.source_transaction.hash())
            writer.write_uint32_le(i.source_output_index)
            script_bin = i.unlocking_script.serialize()
            writer.write_var_int_num(len(script_bin))
            writer.write(script_bin)
            writer.write_uint32_le(i.sequence)
            writer.write_uint64_le(i.source_transaction.outputs[i.source_output_index].satoshis)
            locking_script_bin = i.source_transaction.outputs[i.source_output_index].locking_script.serialize()
            writer.write_var_int_num(len(locking_script_bin))
            writer.write(locking_script_bin)

        writer.write_var_int_num(len(self.outputs))
        for o in self.outputs:
            writer.write_uint64_le(o.satoshis)
            script_bin = o.locking_script.serialize()
            writer.write_var_int_num(len(script_bin))
            writer.write(script_bin)

        writer.write_uint32_le(self.locktime)
        return writer.to_bytes()

    def to_beef(self) -> bytes:
        writer = Writer()
        writer.write_uint32_le(4022206465)
        bumps = []
        txs = []

        def add_paths_and_inputs(tx):
            obj = {'tx': tx}
            has_proof = isinstance(tx.merkle_path, MerklePath)
            if has_proof:
                added = False
                for i, bump in enumerate(bumps):
                    if bump == tx.merkle_path:
                        obj['path_index'] = i
                        added = True
                        break
                    if bump.block_height == tx.merkle_path.block_height:
                        root_a = bump.compute_root()
                        root_b = tx.merkle_path.compute_root()
                        if root_a == root_b:
                            bump.combine(tx.merkle_path)
                            obj['path_index'] = i
                            added = True
                            break
                if not added:
                    obj['path_index'] = len(bumps)
                    bumps.append(tx.merkle_path)
            txs.insert(0, obj)
            if not has_proof:
                for tx_input in tx.inputs:
                    if not isinstance(tx_input.source_transaction, Transaction):
                        raise ValueError('A required source transaction is missing!')
                    add_paths_and_inputs(tx_input.source_transaction)

        add_paths_and_inputs(self)

        writer.write_var_int_num(len(bumps))
        for b in bumps:
            writer.write(b.to_binary())
        writer.write_var_int_num(len(txs))
        for t in txs:
            writer.write(t['tx'].serialize())
            if 'path_index' in t:
                writer.write_uint8(1)
                writer.write_var_int_num(t['path_index'])
            else:
                writer.write_uint8(0)
        return writer.to_bytes()

    @classmethod
    def from_reader(cls, reader: Reader) -> 'Transaction':
        t = cls()
        t.version = reader.read_uint32_le()
        assert t.version is not None
        inputs_count = reader.read_var_int_num()
        assert inputs_count is not None
        for _ in range(inputs_count):
            _input = TransactionInput.from_hex(reader)
            assert _input is not None
            t.inputs.append(_input)
        outputs_count = reader.read_var_int_num()
        assert outputs_count is not None
        for _ in range(outputs_count):
            _output = TransactionOutput.from_hex(reader)
            assert _output is not None
            t.outputs.append(_output)
        t.locktime = reader.read_uint32_le()
        assert t.locktime is not None
        return t

    async def verify(self, chaintracker: Optional[ChainTracker] = default_chain_tracker(), scripts_only=False) -> bool:
        if self.merkle_path and not scripts_only:
            proof_valid = await self.merkle_path.verify(self.txid(), chaintracker)
            if proof_valid:
                return True

        input_total = 0
        for i, tx_input in enumerate(self.inputs):
            if not tx_input.source_transaction:
                raise ValueError(
                    f"Verification failed because the input at index {i} of transaction {self.txid()} "
                    f"is missing an associated source transaction. "
                    f"This source transaction is required for transaction verification because there is no "
                    f"merkle proof for the transaction spending a UTXO it contains.")
            if not tx_input.unlocking_script:
                raise ValueError(
                    f"Verification failed because the input at index {i} of transaction {self.txid()} "
                    f"is missing an associated unlocking script. "
                    f"This script is required for transaction verification because there is no "
                    f"merkle proof for the transaction spending the UTXO.")

            source_output = tx_input.source_transaction.outputs[tx_input.source_output_index]
            input_total += source_output.satoshis

            input_verified = await tx_input.source_transaction.verify(chaintracker)
            if not input_verified:
                return False

            other_inputs = self.inputs[:i] + self.inputs[i + 1:]
            spend = Spend({
                'sourceTXID': tx_input.source_transaction.txid(),
                'sourceOutputIndex': tx_input.source_output_index,
                'sourceSatoshis': source_output.satoshis,
                'lockingScript': source_output.locking_script,
                'transactionVersion': self.version,
                'otherInputs': other_inputs,
                'inputIndex': i,
                'unlockingScript': tx_input.unlocking_script,
                'outputs': self.outputs,
                'inputSequence': tx_input.sequence,
                'lockTime': self.locktime,
            })
            spend_valid = spend.validate()
            if not spend_valid:
                return False

        output_total = 0
        for out in self.outputs:
            if not out.satoshis:
                raise ValueError("Every output must have a defined amount during transaction verification.")
            output_total += out.satoshis

        return output_total <= input_total

    @classmethod
    def parse_script_offsets(cls, octets: Union[bytes, str]) -> Dict[str, List[Dict[str, int]]]:
        """
        Since the validation of blockchain data is atomically transaction data validation,
        any application seeking to validate data in output scripts must store the entire transaction as well.
        Since the transaction data includes the output script data, saving a second copy of potentially
        large scripts can bloat application storage requirements.

        This function efficiently parses binary transaction data to determine the offsets and lengths of each script.
        This supports the efficient retrieval of script data from transaction data.

        @param octets: binary transaction data or hex string
        @returns: {
            inputs: { vin: number, offset: number, length: number }[]
            outputs: { vout: number, offset: number, length: number }[]
        }
        """
        if isinstance(octets, str):
            octets = bytes.fromhex(octets)

        br = Reader(octets)
        inputs: List[Dict[str, int]] = []
        outputs: List[Dict[str, int]] = []

        br.read(4)  # skip version
        inputs_length = br.read_var_int_num()
        for i in range(inputs_length):
            br.read(36)  # skip txid and vout
            script_length = br.read_var_int_num()
            inputs.append({'vin': i, 'offset': br.tell(), 'length': script_length})
            br.read(script_length + 4)  # script and sequence

        outputs_length = br.read_var_int_num()
        for i in range(outputs_length):
            br.read(8)
            script_length = br.read_var_int_num()
            outputs.append({'vout': i, 'offset': br.tell(), 'length': script_length})
            br.read(script_length)  # skip script

        return {'inputs': inputs, 'outputs': outputs}



from abc import ABC, abstractmethod

from .script import Script


class UnlockingScriptTemplate(ABC):

    @staticmethod
    @abstractmethod
    def sign(tx, input_index) -> Script:
        pass

    @staticmethod
    @abstractmethod
    def estimated_unlocking_byte_length() -> int:
        pass



import pytest

from bsv.encrypted_message import EncryptedMessage
from bsv.keys import PrivateKey
from bsv.utils import randbytes


def test_aes_gcm():
    key = randbytes(32)
    message = 'hello world'.encode('utf-8')
    encrypted = EncryptedMessage.aes_gcm_encrypt(key, message)
    decrypted = EncryptedMessage.aes_gcm_decrypt(key, encrypted)
    assert decrypted == message


def test_brc78():
    message = 'hello world'.encode('utf-8')
    sender_priv, recipient_priv = PrivateKey(), PrivateKey()
    encrypted = EncryptedMessage.encrypt(message, sender_priv, recipient_priv.public_key())
    decrypted = EncryptedMessage.decrypt(encrypted, recipient_priv)
    assert decrypted == message

    with pytest.raises(ValueError, match=r'message version mismatch'):
        EncryptedMessage.decrypt(encrypted[1:], PrivateKey())
    with pytest.raises(ValueError, match=r'recipient public key mismatch'):
        EncryptedMessage.decrypt(encrypted, PrivateKey())
    with pytest.raises(ValueError, match=r'failed to decrypt message'):
        EncryptedMessage.decrypt(encrypted[:-1], recipient_priv)



import json
import random
from typing import Optional, Dict, Union, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..transaction import Transaction

from ..broadcaster import BroadcastResponse, BroadcastFailure, Broadcaster
from ..http_client import HttpClient, default_http_client, SyncHttpClient, default_sync_http_client

def to_hex(bytes_data):
    return "".join(f"{x:02x}" for x in bytes_data)


def random_hex(length: int) -> str:
    return "".join(f"{random.randint(0, 255):02x}" for _ in range(length))


class ARCConfig:
    def __init__(
            self,
            api_key: Optional[str] = None,
            http_client: Optional[HttpClient] = None,
            sync_http_client: Optional[SyncHttpClient] = None,
            deployment_id: Optional[str] = None,
            callback_url: Optional[str] = None,
            callback_token: Optional[str] = None,
            headers: Optional[Dict[str, str]] = None,
    ):
        self.api_key = api_key
        self.http_client = http_client
        self.sync_http_client = sync_http_client
        self.deployment_id = deployment_id
        self.callback_url = callback_url
        self.callback_token = callback_token
        self.headers = headers


def default_deployment_id() -> str:
    return f"py-sdk-{random_hex(16)}"


class ARC(Broadcaster):
    def __init__(self, url: str, config: Union[str, ARCConfig] = None):
        self.URL = url
        if isinstance(config, str):
            self.api_key = config
            self.http_client = default_http_client()
            self.sync_http_client = default_sync_http_client()
            self.deployment_id = default_deployment_id()
            self.callback_url = None
            self.callback_token = None
            self.headers = None
        else:
            config = config or ARCConfig()
            self.api_key = config.api_key
            self.http_client = config.http_client or default_http_client()
            self.sync_http_client = config.sync_http_client or default_sync_http_client()
            self.deployment_id = config.deployment_id or default_deployment_id()
            self.callback_url = config.callback_url
            self.callback_token = config.callback_token
            self.headers = config.headers

    async def broadcast(
            self, tx: 'Transaction'
    ) -> Union[BroadcastResponse, BroadcastFailure]:
        # Check if all inputs have source_transaction
        has_all_source_txs = all(input.source_transaction is not None for input in tx.inputs)
        request_options = {
            "method": "POST",
            "headers": self.request_headers(),
            "data": {
                "rawTx":
                    tx.to_ef().hex() if has_all_source_txs else tx.hex()
            }
        }
        try:
            response = await self.http_client.fetch(
                f"{self.URL}/v1/tx", request_options
            )

            response_json = response.json()

            if response.ok and response.status_code >= 200 and response.status_code <= 299:
                data = response_json["data"]

                if data.get("txid"):
                    return BroadcastResponse(
                        status="success",
                        txid=data.get("txid"),
                        message=f"{data.get('txStatus', '')} {data.get('extraInfo', '')}",
                    )
                else:
                    return BroadcastFailure(
                        status="failure",
                        code=data.get("status", "ERR_UNKNOWN"),
                        description=data.get("detail", "Unknown error"),
                    )
            else:
                return BroadcastFailure(
                    status="failure",
                    code=str(response.status_code),
                    description=response_json["data"]["detail"] if "data" in response_json else "Unknown error",
                )

        except Exception as error:
            return BroadcastFailure(
                status="failure",
                code="500",
                description=(
                    str(error)
                    if isinstance(error, Exception)
                    else "Internal Server Error"
                ),
            )

    def request_headers(self) -> Dict[str, str]:
        headers = {
            "Content-Type": "application/json",
            "XDeployment-ID": self.deployment_id,
        }

        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"

        if self.callback_url:
            headers["X-CallbackUrl"] = self.callback_url

        if self.callback_token:
            headers["X-CallbackToken"] = self.callback_token

        if self.headers:
            headers.update(self.headers)

        return headers

    def sync_broadcast(
            self, tx: 'Transaction', timeout: int = 30
    ) -> Union[BroadcastResponse, BroadcastFailure]:
        """
        Synchronously broadcast a transaction

        :param tx: Transaction to broadcast
        :param timeout: Timeout setting in seconds
        :returns: BroadcastResponse or BroadcastFailure
        """
        # Check if all inputs have source_transaction
        has_all_source_txs = all(input.source_transaction is not None for input in tx.inputs)

        try:
            response = self.sync_http_client.post(
                f"{self.URL}/v1/tx",
                data={"rawTx": tx.to_ef().hex() if has_all_source_txs else tx.hex()},
                headers=self.request_headers(),
                timeout=timeout
            )

            response_json = response.json()
            data = response_json.get("data", {})

            if response.ok:
                if data.get("txid"):
                    return BroadcastResponse(
                        status="success",
                        txid=data.get("txid"),
                        message=f"{data.get('txStatus', '')} {data.get('extraInfo', '')}".strip(),
                    )
                else:
                    return BroadcastFailure(
                        status="failure",
                        code=data.get("status", "ERR_UNKNOWN"),
                        description=data.get("detail", "Unknown error"),
                    )
            else:
                # Handle special error cases
                if response.status_code == 408:
                    return BroadcastFailure(
                        status="failure",
                        code="408",
                        description=f"Transaction broadcast timed out after {timeout} seconds",
                    )

                if response.status_code == 503:
                    return BroadcastFailure(
                        status="failure",
                        code="503",
                        description="Failed to connect to ARC service",
                    )

                return BroadcastFailure(
                    status="failure",
                    code=str(response.status_code),
                    description=data.get("detail", "Unknown error"),
                )

        except Exception as error:
            return BroadcastFailure(
                status="failure",
                code="500",
                description=str(error),
            )

    def check_transaction_status(self, txid: str, timeout: int = 5) -> Dict[str, Any]:
        """
        Check transaction status synchronously

        :param txid: Transaction ID to check
        :param timeout: Timeout setting in seconds
        :returns: Dictionary containing transaction status information
        """

        try:
            response = self.sync_http_client.get(
                f"{self.URL}/v1/tx/{txid}",
                headers=self.request_headers(),
                timeout=timeout
            )
            response_data = response.json()
            data = response_data.get("data", {})

            if response.ok:
                return {
                    "txid": txid,
                    "txStatus": data.get("txStatus"),
                    "blockHash": data.get("blockHash"),
                    "blockHeight": data.get("blockHeight"),
                    "merklePath": data.get("merklePath"),
                    "extraInfo": data.get("extraInfo"),
                    "competingTxs": data.get("competingTxs"),
                    "timestamp": data.get("timestamp")
                }
            else:
                # Handle special error cases
                if response.status_code == 408:
                    return {
                        "status": "failure",
                        "code": 408,
                        "title": "Request Timeout",
                        "detail": f"Transaction status check timed out after {timeout} seconds",
                        "txid": txid,
                        "extra_info": "Consider retrying or increasing timeout value"
                    }

                if response.status_code == 503:
                    return {
                        "status": "failure",
                        "code": 503,
                        "title": "Connection Error",
                        "detail": "Failed to connect to ARC service",
                        "txid": txid
                    }

                # Handle general error cases
                return {
                    "status": "failure",
                    "code": data.get("status", response.status_code),
                    "title": data.get("title", "Error"),
                    "detail": data.get("detail", "Unknown error"),
                    "txid": data.get("txid", txid),
                    "extra_info": data.get("extraInfo", "")
                }

        except Exception as error:
            return {
                "status": "failure",
                "code": "500",
                "title": "Internal Error",
                "detail": str(error),
                "txid": txid
            }

    @staticmethod
    def categorize_transaction_status(response: Dict[str, Any]) -> Dict[str, Any]:
        """
        Categorize transaction status based on the ARC response

        :param response: The transaction status response dictionary from ARC
        :returns: Dictionary containing status category and transaction status
        """
        try:
            tx_status = response.get("txStatus")

            if tx_status:
                # Processing transactions - still being handled by the network
                if tx_status in [
                    "UNKNOWN", "QUEUED", "RECEIVED", "STORED",
                    "ANNOUNCED_TO_NETWORK", "REQUESTED_BY_NETWORK",
                    "SENT_TO_NETWORK", "ACCEPTED_BY_NETWORK"
                ]:
                    status_category = "progressing"

                # Successfully mined transactions
                elif tx_status in ["MINED"]:
                    status_category = "mined"

                # Mined in stale block - needs attention
                elif tx_status in ["MINED_IN_STALE_BLOCK"]:
                    status_category = "0confirmation"

                # Warning status - double spend attempted
                elif tx_status in ["DOUBLE_SPEND_ATTEMPTED"]:
                    status_category = "warning"

                # Seen on network - check for competing transactions
                elif tx_status in ["SEEN_ON_NETWORK"]:
                    # Check if there are competing transactions in mempool
                    if response.get("competingTxs"):
                        status_category = "warning"
                    else:
                        # Transaction is in mempool without conflicts
                        status_category = "0confirmation"

                # Rejected transactions - failed to process
                elif tx_status in ["ERROR", "REJECTED", "SEEN_IN_ORPHAN_MEMPOOL"]:
                    status_category = "rejected"

                else:
                    status_category = f"unknown_txStatus: {tx_status}"
            else:
                status_category = "error"
                tx_status = "No txStatus"

            return {
                "status_category": status_category,
                "tx_status": tx_status
            }

        except Exception as e:
            return {
                "status_category": "error",
                "error": str(e),
                "response": response
            }


from io import BytesIO
from typing import List

from .constants import SIGHASH
from .hash import hash256
from .transaction_input import TransactionInput
from .transaction_output import TransactionOutput


def _preimage(
        tx_input: TransactionInput,
        tx_version: int,
        tx_locktime: int,
        hash_prevouts: bytes,
        hash_sequence: bytes,
        hash_outputs: bytes,
) -> bytes:
    """
    BIP-143 https://github.com/bitcoin/bips/blob/master/bip-0143.mediawiki
     1. nVersion of the transaction (4-byte little endian)
     2. hashPrevouts (32-byte hash)
     3. hashSequence (32-byte hash)
     4. outpoint (32-byte hash + 4-byte little endian)
     5. scriptCode of the input (serialized as scripts inside CTxOuts)
     6. value of the output spent by this input (8-byte little endian)
     7. nSequence of the input (4-byte little endian)
     8. hashOutputs (32-byte hash)
     9. nLocktime of the transaction (4-byte little endian)
    10. sighash type of the signature (4-byte little endian)
    """
    stream = BytesIO()
    # 1
    stream.write(tx_version.to_bytes(4, "little"))
    # 2
    stream.write(hash_prevouts)
    # 3
    stream.write(hash_sequence)
    # 4
    stream.write(bytes.fromhex(tx_input.source_txid)[::-1])
    stream.write(tx_input.source_output_index.to_bytes(4, "little"))
    # 5
    stream.write(tx_input.locking_script.byte_length_varint())
    stream.write(tx_input.locking_script.serialize())
    # 6
    stream.write(tx_input.satoshis.to_bytes(8, "little"))
    # 7
    stream.write(tx_input.sequence.to_bytes(4, "little"))
    # 8
    stream.write(hash_outputs)
    # 9
    stream.write(tx_locktime.to_bytes(4, "little"))
    # 10
    stream.write(tx_input.sighash.to_bytes(4, "little"))
    return stream.getvalue()


def tx_preimages(
        inputs: List[TransactionInput],
        outputs: List[TransactionOutput],
        tx_version: int,
        tx_locktime: int,
) -> List[bytes]:
    """
    :returns: the digests of unsigned transaction
    """
    _hash_prevouts = hash256(
        b"".join(
            bytes.fromhex(_in.source_txid)[::-1] + _in.source_output_index.to_bytes(4, "little")
            for _in in inputs
        )
    )
    _hash_sequence = hash256(
        b"".join(_in.sequence.to_bytes(4, "little") for _in in inputs)
    )
    _hash_outputs = hash256(
        b"".join(tx_output.serialize() for tx_output in outputs)
    )
    digests = []
    for i in range(len(inputs)):
        sighash = inputs[i].sighash
        # hash previous outs
        if not sighash & SIGHASH.ANYONECANPAY:
            # if anyone can pay is not set
            hash_prevouts = _hash_prevouts
        else:
            hash_prevouts = b"\x00" * 32
        # hash sequence
        if (
                not sighash & SIGHASH.ANYONECANPAY
                and sighash & 0x1F != SIGHASH.SINGLE
                and sighash & 0x1F != SIGHASH.NONE
        ):
            # if none of anyone can pay, single, none is set
            hash_sequence = _hash_sequence
        else:
            hash_sequence = b"\x00" * 32
        # hash outputs
        if sighash & 0x1F != SIGHASH.SINGLE and sighash & 0x1F != SIGHASH.NONE:
            # if neither single nor none
            hash_outputs = _hash_outputs
        elif sighash & 0x1F == SIGHASH.SINGLE and i < len(outputs):
            # if single and the input index is smaller than the number of outputs
            hash_outputs = hash256(outputs[i].serialize())
        else:
            hash_outputs = b"\x00" * 32
        digests.append(
            _preimage(inputs[i], tx_version, tx_locktime, hash_prevouts, hash_sequence, hash_outputs)
        )
    return digests


def tx_preimage(
        input_index: int,
        inputs: List[TransactionInput],
        outputs: List[TransactionOutput],
        tx_version: int,
        tx_locktime: int,
) -> bytes:
    """
    Calculates and returns the preimage for a specific input index.
    """
    sighash = inputs[input_index].sighash

    # hash previous outs
    if not sighash & SIGHASH.ANYONECANPAY:
        hash_prevouts = hash256(
            b"".join(
                bytes.fromhex(_in.source_txid)[::-1] + _in.source_output_index.to_bytes(4, "little")
                for _in in inputs
            )
        )
    else:
        hash_prevouts = b"\x00" * 32

    # hash sequence
    if (
            not sighash & SIGHASH.ANYONECANPAY
            and sighash & 0x1F != SIGHASH.SINGLE
            and sighash & 0x1F != SIGHASH.NONE
    ):
        hash_sequence = hash256(
            b"".join(_in.sequence.to_bytes(4, "little") for _in in inputs)
        )
    else:
        hash_sequence = b"\x00" * 32

    # hash outputs
    if sighash & 0x1F != SIGHASH.SINGLE and sighash & 0x1F != SIGHASH.NONE:
        hash_outputs = hash256(
            b"".join(tx_output.serialize() for tx_output in outputs)
        )
    elif sighash & 0x1F == SIGHASH.SINGLE and input_index < len(outputs):
        hash_outputs = hash256(outputs[input_index].serialize())
    else:
        hash_outputs = b"\x00" * 32

    return _preimage(inputs[input_index], tx_version, tx_locktime, hash_prevouts, hash_sequence, hash_outputs)



import secrets
from typing import List, Optional, Union

from .base58 import b58_encode, b58_decode
from .curve import curve


# Define PointInFiniteField class for points in finite field
class PointInFiniteField:
    """
    A class representing a point in a finite field.

    Used for implementing Shamir's Secret Sharing scheme.
    """

    def __init__(self, x: Union[int, bytes], y: Union[int, bytes]):
        """
        Initialize a point in a finite field.

        Args:
            x: The x-coordinate (can be int or bytes)
            y: The y-coordinate (can be int or bytes)
        """
        # Convert inputs to integers if they are bytes
        if isinstance(x, bytes):
            x = int.from_bytes(x, 'big')
        if isinstance(y, bytes):
            y = int.from_bytes(y, 'big')

        # Take modulo of coordinates with prime field size
        self.x = x % curve.p
        self.y = y % curve.p

    def __str__(self) -> str:
        """
        String representation of the point using base58 encoding.

        Returns:
            String in format x.y where x and y are base58 encoded
        """
        # Convert integers to bytes and then base58 encode
        x_bytes = self.x.to_bytes((self.x.bit_length() + 7) // 8, 'big')
        y_bytes = self.y.to_bytes((self.y.bit_length() + 7) // 8, 'big')

        return f"{b58_encode(x_bytes)}.{b58_encode(y_bytes)}"

    @staticmethod
    def from_string(s: str) -> 'PointInFiniteField':
        """
        Create a point from its string representation.

        Args:
            s: String in format x.y where x and y are base58 encoded

        Returns:
            A PointInFiniteField object

        Raises:
            ValueError: If the string is not in the correct format
        """
        parts = s.split('.')
        if len(parts) < 2:
            raise ValueError(f"Invalid point format: {s}")

        x, y = parts[0], parts[1]

        # Decode base58 strings to bytes, then to integers
        x_int = int.from_bytes(b58_decode(x), 'big')
        y_int = int.from_bytes(b58_decode(y), 'big')

        return PointInFiniteField(x_int, y_int)


class Polynomial:
    """
    A class representing a polynomial for Shamir's Secret Sharing.

    This is used to split a private key into shares such that a certain
    threshold of shares is required to reconstruct the key.
    """

    def __init__(self, points: List[PointInFiniteField], threshold: Optional[int] = None):
        """
        Initialize a polynomial with the given points.

        Args:
            points: List of points defining the polynomial
            threshold: Number of points required to reconstruct the polynomial.
                      If None, uses the length of points.
        """
        self.points = points
        # Default threshold is the number of points if not specified
        self.threshold = threshold if threshold is not None else len(points)

    @staticmethod
    def from_private_key(private_key_int: int, threshold: int) -> 'Polynomial':
        """
        Create a polynomial from a private key.

        The private key becomes the y-intercept (at x=0) of the polynomial.
        Additional random points are generated to define the polynomial.

        Args:
            private_key_int: The private key (integer))
            threshold: Number of shares required to reconstruct the key

        Returns:
            A Polynomial object
        """

        # Create first point at x=0, y=private_key
        points = [PointInFiniteField(0, private_key_int)]

        # Generate additional random coefficients (as points)
        for i in range(1, threshold):
            # Generate cryptographically secure random values for x and y
            # The random function from secrets should be more secure than random
            random_x_bytes = secrets.token_bytes(32)
            random_y_bytes = secrets.token_bytes(32)

            # Convert to integers and take modulo p
            random_x = int.from_bytes(random_x_bytes, 'big') % curve.p
            random_y = int.from_bytes(random_y_bytes, 'big') % curve.p

            points.append(PointInFiniteField(random_x, random_y))

        return Polynomial(points)

    def value_at(self, x: Union[int, bytes]) -> int:
        """
        Evaluate the polynomial at the given x coordinate using Lagrange interpolation.

        Args:
            x: The x coordinate at which to evaluate the polynomial

        Returns:
            The y coordinate (as integer) corresponding to the given x
        """
        # Convert x to integer if it's bytes
        if isinstance(x, bytes):
            x = int.from_bytes(x, 'big')

        # Ensure x is within the field
        x = x % curve.p

        # Initialize result
        y = 0

        # Lagrange interpolation
        for i in range(self.threshold):
            term = self.points[i].y
            for j in range(self.threshold):
                if i != j:
                    xi = self.points[i].x
                    xj = self.points[j].x

                    # Calculate numerator: (x - xj) mod p
                    numerator = (x - xj) % curve.p

                    # Calculate denominator: (xi - xj) mod p
                    denominator = (xi - xj) % curve.p

                    # Calculate modular inverse of denominator
                    # pow(base, exponent, modulus) computes (base^exponent) % modulus
                    # When exponent is -1, it computes the modular multiplicative inverse
                    denominator_inverse = pow(denominator, -1, curve.p)

                    # Calculate the fraction: numerator * denominator_inverse mod p
                    fraction = (numerator * denominator_inverse) % curve.p

                    # Multiply the current term by the fraction
                    term = (term * fraction) % curve.p

            # Add this term to the result
            y = (y + term) % curve.p

        return y


class KeyShares:
    """
    A class representing key shares for Shamir's Secret Sharing.

    This is used to store the shares of a split private key along with
    metadata like threshold and integrity hash.
    """

    def __init__(self, points: List[PointInFiniteField], threshold: int, integrity: str):
        """
        Initialize key shares.

        Args:
            points: List of points representing the shares
            threshold: Number of shares required to reconstruct the key
            integrity: Integrity check hash derived from the public key
        """
        self.points = points
        self.threshold = threshold
        self.integrity = integrity

    @staticmethod
    def from_backup_format(shares: List[str]) -> 'KeyShares':
        """
        Create KeyShares from backup format strings.

        Args:
            shares: List of share strings in format "x.y.t.i"

        Returns:
            A KeyShares object

        Raises:
            ValueError: If shares have invalid format or inconsistent threshold/integrity
        """
        threshold = 0
        integrity = ''
        points = []

        for idx, share in enumerate(shares):
            # Split the share string into parts
            share_parts = share.split('.')
            if len(share_parts) != 4:
                raise ValueError(
                    f'Invalid share format in share {idx}. '
                    f'Expected format: "x.y.t.i" - received {share}'
                )

            # Parse the parts
            x, y, t, i = share_parts

            if not t:
                raise ValueError(f'Threshold not found in share {idx}')
            if not i:
                raise ValueError(f'Integrity not found in share {idx}')

            # Parse threshold as integer
            t_int = int(t)

            # Check consistency across shares
            if idx > 0 and threshold != t_int:
                raise ValueError(f'Threshold mismatch in share {idx}')
            if idx > 0 and integrity != i:
                raise ValueError(f'Integrity mismatch in share {idx}')

            threshold = t_int
            integrity = i

            # Create point from x and y components
            point = PointInFiniteField.from_string(f"{x}.{y}")
            points.append(point)

        return KeyShares(points, threshold, integrity)

    def to_backup_format(self) -> List[str]:
        """
        Convert shares to backup format strings.

        Returns:
            List of share strings in format "x.y.t.i"
        """
        return [
            f"{point}.{self.threshold}.{self.integrity}"
            for point in map(str, self.points)
        ]


import math
import re
import struct
from base64 import b64encode, b64decode
from contextlib import suppress
from io import BytesIO
from secrets import randbits
from typing import Tuple, Optional, Union, Literal, List

from .base58 import base58check_decode
from .constants import Network, ADDRESS_PREFIX_NETWORK_DICT, WIF_PREFIX_NETWORK_DICT, NUMBER_BYTE_LENGTH
from .constants import OpCode
from .curve import curve


def unsigned_to_varint(num: int) -> bytes:
    """
    convert an unsigned int to varint.
    """
    if num < 0 or num > 0xffffffffffffffff:
        raise OverflowError(f"can't convert {num} to varint")
    if num <= 0xfc:
        return num.to_bytes(1, 'little')
    elif num <= 0xffff:
        return b'\xfd' + num.to_bytes(2, 'little')
    elif num <= 0xffffffff:
        return b'\xfe' + num.to_bytes(4, 'little')
    else:
        return b'\xff' + num.to_bytes(8, 'little')


def unsigned_to_bytes(num: int, byteorder: Literal['big', 'little'] = 'big') -> bytes:
    """
    convert an unsigned int to the least number of bytes as possible.
    """
    return num.to_bytes(math.ceil(num.bit_length() / 8) or 1, byteorder)


def decode_address(address: str) -> Tuple[bytes, Network]:
    """
    :returns: tuple (public_key_hash_bytes, network)
    """
    if not re.match(r'^[1mn][a-km-zA-HJ-NP-Z1-9]{24,33}$', address):
        # - a Bitcoin address is between 25 and 34 characters long;
        # - the address always starts with a 1, m, or n
        # - an address can contain all alphanumeric characters, with the exceptions of 0, O, I, and l.
        raise ValueError(f'invalid P2PKH address {address}')
    decoded = base58check_decode(address)
    prefix = decoded[:1]
    network = ADDRESS_PREFIX_NETWORK_DICT.get(prefix)
    return decoded[1:], network


def validate_address(address: str, network: Optional[Network] = None) -> bool:
    """
    :returns: True if address is a valid bitcoin legacy address (P2PKH)
    """
    with suppress(Exception):
        _, _network = decode_address(address)
        if network is not None:
            return _network == network
        return True
    return False


def address_to_public_key_hash(address: str) -> bytes:
    """
    :returns: convert P2PKH address to the corresponding public key hash
    """
    return decode_address(address)[0]


def decode_wif(wif: str) -> Tuple[bytes, bool, Network]:
    """
    :returns: tuple (private_key_bytes, compressed, network)
    """
    decoded = base58check_decode(wif)
    prefix = decoded[:1]
    network = WIF_PREFIX_NETWORK_DICT.get(prefix)
    if not network:
        raise ValueError(f'unknown WIF prefix {prefix.hex()}')
    if len(wif) == 52 and decoded[-1] == 1:
        return decoded[1:-1], True, network
    return decoded[1:], False, network


def deserialize_ecdsa_der(signature: bytes) -> Tuple[int, int]:
    """
    deserialize ECDSA signature from bitcoin strict DER to (r, s)
    """
    try:
        assert signature[0] == 0x30
        assert int(signature[1]) == len(signature) - 2
        # r
        assert signature[2] == 0x02
        r_len = int(signature[3])
        r = int.from_bytes(signature[4: 4 + r_len], 'big')
        # s
        assert signature[4 + r_len] == 0x02
        s_len = int(signature[5 + r_len])
        s = int.from_bytes(signature[-s_len:], 'big')
        return r, s
    except Exception:
        raise ValueError(f'invalid DER encoded {signature.hex()}')


def serialize_ecdsa_der(signature: Tuple[int, int]) -> bytes:
    """
    serialize ECDSA signature (r, s) to bitcoin strict DER format
    """
    r, s = signature
    # enforce low s value
    if s > curve.n // 2:
        s = curve.n - s
    # r
    r_bytes = r.to_bytes(NUMBER_BYTE_LENGTH, 'big').lstrip(b'\x00')
    if r_bytes[0] & 0x80:
        r_bytes = b'\x00' + r_bytes
    serialized = bytes([2, len(r_bytes)]) + r_bytes
    # s
    s_bytes = s.to_bytes(NUMBER_BYTE_LENGTH, 'big').lstrip(b'\x00')
    if s_bytes[0] & 0x80:
        s_bytes = b'\x00' + s_bytes
    serialized += bytes([2, len(s_bytes)]) + s_bytes
    return bytes([0x30, len(serialized)]) + serialized


def deserialize_ecdsa_recoverable(signature: bytes) -> Tuple[int, int, int]:
    """
    deserialize recoverable ECDSA signature from bytes to (r, s, recovery_id)
    """
    assert len(signature) == 65, 'invalid length of recoverable ECDSA signature'
    rec_id = signature[-1]
    assert 0 <= rec_id <= 3, f'invalid recovery id {rec_id}'
    r = int.from_bytes(signature[:NUMBER_BYTE_LENGTH], 'big')
    s = int.from_bytes(signature[NUMBER_BYTE_LENGTH:-1], 'big')
    return r, s, rec_id


def serialize_ecdsa_recoverable(signature: Tuple[int, int, int]) -> bytes:
    """
    serialize recoverable ECDSA signature from (r, s, recovery_id) to bytes
    """
    _r, _s, _rec_id = signature
    assert 0 <= _rec_id < 4, f'invalid recovery id {_rec_id}'
    r = _r.to_bytes(NUMBER_BYTE_LENGTH, 'big')
    s = _s.to_bytes(NUMBER_BYTE_LENGTH, 'big')
    rec_id = _rec_id.to_bytes(1, 'big')
    return r + s + rec_id


def serialize_text(text: str) -> bytes:
    """
    serialize plain text to bytes in format: varint_length + text.utf-8
    """
    message: bytes = text.encode('utf-8')
    return unsigned_to_varint(len(message)) + message


def text_digest(text: str) -> bytes:
    """
    :returns: the digest of arbitrary text when signing with bitcoin private key
    """
    return serialize_text('Bitcoin Signed Message:\n') + serialize_text(text)


def stringify_ecdsa_recoverable(signature: bytes, compressed: bool = True) -> str:
    """stringify serialize recoverable ECDSA signature
    :param signature: serialized recoverable ECDSA signature in "r (32 bytes) + s (32 bytes) + recovery_id (1 byte)"
    :param compressed: True if used compressed public key
    :returns: stringified recoverable signature formatted in base64
    """
    r, s, recovery_id = deserialize_ecdsa_recoverable(signature)
    prefix: int = 27 + recovery_id + (4 if compressed else 0)
    signature: bytes = prefix.to_bytes(1, 'big') + signature[:-1]
    return b64encode(signature).decode('ascii')


def unstringify_ecdsa_recoverable(signature: str) -> Tuple[bytes, bool]:
    """
    :returns: (serialized_recoverable_signature, used_compressed_public_key)
    """
    serialized = b64decode(signature)
    assert len(serialized) == 65, 'invalid length of recoverable ECDSA signature'
    prefix = serialized[0]
    assert 27 <= prefix < 35, f'invalid recoverable ECDSA signature prefix {prefix}'
    compressed = False
    if prefix >= 31:
        compressed = True
        prefix -= 4
    recovery_id = prefix - 27
    return serialized[1:] + recovery_id.to_bytes(1, 'big'), compressed


def bytes_to_bits(octets: Union[str, bytes]) -> str:
    """
    convert bytes to binary 0/1 string
    """
    b: bytes = octets if isinstance(octets, bytes) else bytes.fromhex(octets)
    bits: str = bin(int.from_bytes(b, 'big'))[2:]
    if len(bits) < len(b) * 8:
        bits = '0' * (len(b) * 8 - len(bits)) + bits
    return bits


def bits_to_bytes(bits: str) -> bytes:
    """
    convert binary 0/1 string to bytes
    """
    byte_length = math.ceil(len(bits) / 8) or 1
    return int(bits, 2).to_bytes(byte_length, byteorder='big')


def randbytes(length: int) -> bytes:
    """
    generate cryptographically secure random bytes
    """
    return randbits(length * 8).to_bytes(length, 'big')


def get_pushdata_code(byte_length: int) -> bytes:
    """
    :returns: the corresponding PUSHDATA opcode according to the byte length of pushdata
    """
    if byte_length <= 0x4b:
        return byte_length.to_bytes(1, 'little')
    elif byte_length <= 0xff:
        # OP_PUSHDATA1
        return OpCode.OP_PUSHDATA1 + byte_length.to_bytes(1, 'little')
    elif byte_length <= 0xffff:
        # OP_PUSHDATA2
        return OpCode.OP_PUSHDATA2 + byte_length.to_bytes(2, 'little')
    elif byte_length <= 0xffffffff:
        # OP_PUSHDATA4
        return OpCode.OP_PUSHDATA4 + byte_length.to_bytes(4, 'little')
    else:
        raise ValueError("data too long to encode in a PUSHDATA opcode")


def encode_pushdata(pushdata: bytes, minimal_push: bool = True) -> bytes:
    """encode pushdata with proper opcode
    https://github.com/bitcoin-sv/bitcoin-sv/blob/v1.0.10/src/script/interpreter.cpp#L310-L337
    :param pushdata: bytes you want to push onto the stack in bitcoin script
    :param minimal_push: if True then push data following the minimal push rule
    """
    if minimal_push:
        if pushdata == b'':
            return OpCode.OP_0
        if len(pushdata) == 1 and 1 <= pushdata[0] <= 16:
            return bytes([OpCode.OP_1[0] + pushdata[0] - 1])
        if len(pushdata) == 1 and pushdata[0] == 0x81:
            return OpCode.OP_1NEGATE
    else:
        # non-minimal push requires pushdata != b''
        assert pushdata, 'empty pushdata'
    return get_pushdata_code(len(pushdata)) + pushdata


def encode_int(num: int) -> bytes:
    """
    encode a signed integer you want to push onto the stack in bitcoin script, following the minimal push rule
    """
    if num == 0:
        return OpCode.OP_0
    negative: bool = num < 0
    octets: bytearray = bytearray(unsigned_to_bytes(-num if negative else num, 'little'))
    if octets[-1] & 0x80:
        octets += b'\x00'
    if negative:
        octets[-1] |= 0x80
    return encode_pushdata(octets)


def to_hex(byte_array: bytes) -> str:
    return byte_array.hex()


def to_bytes(msg: Union[bytes, str], enc: Optional[str] = None) -> bytes:
    """Converts various message formats into a bytes object."""
    if isinstance(msg, bytes):
        return msg

    if not msg:
        return bytes()

    if isinstance(msg, str):
        if enc == 'hex':
            msg = ''.join(filter(str.isalnum, msg))
            if len(msg) % 2 != 0:
                msg = '0' + msg
            return bytes(int(msg[i:i + 2], 16) for i in range(0, len(msg), 2))
        elif enc == 'base64':
            import base64
            return base64.b64decode(msg)
        else:  # UTF-8 encoding
            return msg.encode('utf-8')

    return bytes(msg)


def to_utf8(arr: List[int]) -> str:
    """Converts an array of numbers to a UTF-8 encoded string."""
    return bytes(arr).decode('utf-8')


def encode(arr: List[int], enc: Optional[str] = None) -> Union[str, List[int]]:
    """Encodes an array of numbers into a specified encoding ('hex' or 'utf8')."""
    if enc == 'hex':
        return to_hex(bytes(arr))
    elif enc == 'utf8':
        return to_utf8(arr)
    return arr


def to_base64(byte_array: List[int]) -> str:
    """Converts an array of bytes into a base64 encoded string."""
    import base64
    return base64.b64encode(bytes(byte_array)).decode('ascii')


base58chars = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def from_base58(str_: str) -> List[int]:
    """Converts a base58 string to a binary array."""
    if not str_ or not isinstance(str_, str):
        raise ValueError(f"Expected base58 string but got '{str_}'")
    if '0' in str_ or 'I' in str_ or 'O' in str_ or 'l' in str_:
        raise ValueError(f"Invalid base58 character in '{str_}'")

    lz = len(str_) - len(str_.lstrip('1'))
    psz = lz

    acc = 0
    for char in str_:
        acc = acc * 58 + base58chars.index(char)

    result = []
    while acc > 0:
        result.append(acc % 256)
        acc //= 256

    return [0] * psz + list(reversed(result))


def to_base58(bin_: List[int]) -> str:
    """Converts a binary array into a base58 string."""
    acc = 0
    for byte in bin_:
        acc = acc * 256 + byte

    result = ''
    while acc > 0:
        acc, mod = divmod(acc, 58)
        result = base58chars[mod] + result

    for byte in bin_:
        if byte == 0:
            result = '1' + result
        else:
            break

    return result


def to_base58_check(bin_: List[int], prefix: Optional[List[int]] = None) -> str:
    """Converts a binary array into a base58check string with a checksum."""
    import hashlib
    if prefix is None:
        prefix = [0]
    hash_ = hashlib.sha256(hashlib.sha256(bytes(prefix + bin_)).digest()).digest()
    return to_base58(prefix + bin_ + list(hash_[:4]))


def from_base58_check(str_: str, enc: Optional[str] = None, prefix_length: int = 1):
    """Converts a base58check string into a binary array after validating the checksum."""
    bin_ = from_base58(str_)
    prefix = bin_[:prefix_length]
    data = bin_[prefix_length:-4]
    checksum = bin_[-4:]

    import hashlib
    hash_ = hashlib.sha256(hashlib.sha256(bytes(prefix + data)).digest()).digest()
    if list(hash_[:4]) != checksum:
        raise ValueError('Invalid checksum')

    if enc == 'hex':
        prefix = to_hex(bytes(prefix))
        data = to_hex(bytes(data))

    return {'prefix': prefix, 'data': data}


class Writer(BytesIO):
    def __init__(self):
        super().__init__()

    def write(self, buf: bytes) -> 'Writer':
        super().write(buf)
        return self

    def write_reverse(self, buf: bytes) -> 'Writer':
        super().write(buf[::-1])
        return self

    def write_uint8(self, n: int) -> 'Writer':
        self.write(struct.pack('B', n))
        return self

    def write_int8(self, n: int) -> 'Writer':
        self.write(struct.pack('b', n))
        return self

    def write_uint16_be(self, n: int) -> 'Writer':
        self.write(struct.pack('>H', n))
        return self

    def write_int16_be(self, n: int) -> 'Writer':
        self.write(struct.pack('>h', n))
        return self

    def write_uint16_le(self, n: int) -> 'Writer':
        self.write(struct.pack('<H', n))
        return self

    def write_int16_le(self, n: int) -> 'Writer':
        self.write(struct.pack('<h', n))
        return self

    def write_uint32_be(self, n: int) -> 'Writer':
        self.write(struct.pack('>I', n))
        return self

    def write_int32_be(self, n: int) -> 'Writer':
        self.write(struct.pack('>i', n))
        return self

    def write_uint32_le(self, n: int) -> 'Writer':
        self.write(struct.pack('<I', n))
        return self

    def write_int32_le(self, n: int) -> 'Writer':
        self.write(struct.pack('<i', n))
        return self

    def write_uint64_be(self, n: int) -> 'Writer':
        self.write(struct.pack('>Q', n))
        return self

    def write_uint64_le(self, n: int) -> 'Writer':
        self.write(struct.pack('<Q', n))
        return self

    def write_var_int_num(self, n: int) -> 'Writer':
        self.write(self.var_int_num(n))
        return self

    def to_bytes(self) -> bytes:
        return self.getvalue()

    @staticmethod
    def var_int_num(n: int) -> bytes:
        return unsigned_to_varint(n)


class Reader(BytesIO):
    def __init__(self, data: bytes):
        super().__init__(data)

    def eof(self) -> bool:
        return self.tell() >= len(self.getvalue())

    def read(self, length: int = None) -> bytes:
        result = super().read(length)
        return result if result else None

    def read_reverse(self, length: int = None) -> bytes:
        data = self.read(length)
        return data[::-1] if data else None

    def read_uint8(self) -> Optional[int]:
        data = self.read(1)
        return data[0] if data else None

    def read_int8(self) -> Optional[int]:
        data = self.read(1)
        return int.from_bytes(data, byteorder='big', signed=True) if data else None

    def read_uint16_be(self) -> Optional[int]:
        data = self.read(2)
        return int.from_bytes(data, byteorder='big') if data else None

    def read_int16_be(self) -> Optional[int]:
        data = self.read(2)
        return int.from_bytes(data, byteorder='big', signed=True) if data else None

    def read_uint16_le(self) -> Optional[int]:
        data = self.read(2)
        return int.from_bytes(data, byteorder='little') if data else None

    def read_int16_le(self) -> Optional[int]:
        data = self.read(2)
        return int.from_bytes(data, byteorder='little', signed=True) if data else None

    def read_uint32_be(self) -> Optional[int]:
        data = self.read(4)
        return int.from_bytes(data, byteorder='big') if data else None

    def read_int32_be(self) -> Optional[int]:
        data = self.read(4)
        return int.from_bytes(data, byteorder='big', signed=True) if data else None

    def read_uint32_le(self) -> Optional[int]:
        data = self.read(4)
        return int.from_bytes(data, byteorder='little') if data else None

    def read_int32_le(self) -> Optional[int]:
        data = self.read(4)
        return int.from_bytes(data, byteorder='little', signed=True) if data else None

    def read_var_int_num(self) -> Optional[int]:
        first_byte = self.read_uint8()
        if first_byte is None:
            return None
        if first_byte < 253:
            return first_byte
        elif first_byte == 253:
            return self.read_uint16_le()
        elif first_byte == 254:
            return self.read_uint32_le()
        elif first_byte == 255:
            data = self.read(8)
            return int.from_bytes(data, byteorder='little') if data else None
        else:
            raise ValueError("Invalid varint encoding")

    def read_var_int(self) -> Optional[bytes]:
        first_byte = self.read(1)
        if not first_byte:
            return None
        if first_byte[0] == 0xfd:
            return first_byte + (self.read(2) or b'')
        elif first_byte[0] == 0xfe:
            return first_byte + (self.read(4) or b'')
        elif first_byte[0] == 0xff:
            return first_byte + (self.read(8) or b'')
        else:
            return first_byte

    def read_bytes(self, byte_length: Optional[int] = None) -> bytes:
        result = self.read(byte_length)
        return result if result else b''

    def read_int(
            self, byte_length: int, byteorder: Literal["big", "little"] = "little"
    ) -> Optional[int]:
        octets = self.read_bytes(byte_length)
        if not octets:
            return None
        return int.from_bytes(octets, byteorder=byteorder)
    
    
def reverse_hex_byte_order(hex_str: str):
    return bytes.fromhex(hex_str)[::-1].hex()



from bsv.hash import sha256, double_sha256, ripemd160_sha256, hmac_sha256, hmac_sha512

MESSAGE = 'hello'.encode('utf-8')
MESSAGE_SHA256 = bytes.fromhex('2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824')
MESSAGE_HASH256 = bytes.fromhex('9595c9df90075148eb06860365df33584b75bff782a510c6cd4883a419833d50')
MESSAGE_HASH160 = bytes.fromhex('b6a9c8c230722b7c748331a8b450f05566dc7d0f')


def test_sha256():
    assert sha256(MESSAGE) == MESSAGE_SHA256


def test_double_sha256():
    assert double_sha256(MESSAGE) == MESSAGE_HASH256


def test_ripemd160_sha256():
    assert ripemd160_sha256(MESSAGE) == MESSAGE_HASH160


KEY = 'key'.encode('utf-8')
MESSAGE_HMAC_SHA256 = bytes.fromhex('9307b3b915efb5171ff14d8cb55fbcc798c6c0ef1456d66ded1a6aa723a58b7b')
MESSAGE_HMAC_SHA512 = bytes.fromhex('ff06ab36757777815c008d32c8e14a705b4e7bf310351a06a23b612dc4c7433e\
                                     7757d20525a5593b71020ea2ee162d2311b247e9855862b270122419652c0c92')


def test_hmac_sha256():
    assert hmac_sha256(KEY, MESSAGE) == MESSAGE_HMAC_SHA256


def test_hmac_sha512():
    assert hmac_sha512(KEY, MESSAGE) == MESSAGE_HMAC_SHA512



from typing import Optional, Dict

from ..chaintracker import ChainTracker
from ..http_client import HttpClient, default_http_client


class WhatsOnChainTracker(ChainTracker):
    def __init__(
            self,
            network: str = "main",
            api_key: Optional[str] = None,
            http_client: Optional[HttpClient] = None,
    ):
        self.network = network
        self.URL = f"https://api.whatsonchain.com/v1/bsv/{network}"
        self.http_client = (
            http_client if http_client else default_http_client()
        )
        self.api_key = api_key

    async def is_valid_root_for_height(self, root: str, height: int) -> bool:
        request_options = {"method": "GET", "headers": self.get_headers()}

        response = await self.http_client.fetch(
            f"{self.URL}/block/{height}/header", request_options
        )
        if response.ok:
            merkleroot = response.json()['data'].get("merkleroot")
            return merkleroot == root
        elif response.status_code == 404:
            return False
        else:
            raise Exception(
                f"Failed to verify merkleroot for height {height} because of an error: {response.json()}"
            )

    def get_headers(self) -> Dict[str, str]:
        headers = {}
        if self.api_key:
            headers["Authorization"] = self.api_key
        return headers
